from . import index
from .index import FLANNParameters, FLANNIndex, set_distance_type

try:
    from numpy.testing import nosetester
    test = nosetester.NoseTester().test
    del nosetester
except ImportError:
    pass

_flann_lib = None
_ldd_re = r'^\t([^\s]+) => (.*) \(0x[\da-f]+\)$'
def get_flann_lib():
    '''
    Gets the location of the flann library that users should link against.
    '''
    # Would make sense to have setup.py write out a file containing the location
    # it found (like what CMake does), but that doesn't work with Anaconda envs.
    global _flann_lib
    if _flann_lib is not None:
        return _flann_lib

    import os
    import re
    import subprocess
    import sys

    so_name = index.__file__

    if sys.platform == 'darwin':
        out = subprocess.check_output(['otool', '-L', so_name]).split('\n')
        assert out.pop(0) == so_name + ':'
        for line in out:
            assert line[0] == '\t'
            fname = line.split(None, 1)[0]
            if 'libflann' in fname and 'libflann_cpp' not in fname:
                fname = fname.replace('@loader_path', os.path.dirname(so_name))
                if '/' not in fname:
                    # relative install_names, dammit
                    # look for it in the standard places...
                    dirs = ['/usr/local/lib', '/usr/lib', '/lib']
                    if 'LIBRARY_PATH' in os.environ:
                        dirs = os.environ.split(':') + dirs

                    from distutils.unixccompiler import UnixCCompiler
                    assert fname.startswith('lib')
                    assert fname.endswith('.dylib')
                    basename = fname[len('lib'):-len('.dylib')]
                    fname = UnixCCompiler().find_library_file(dirs, basename)
                    if fname is None:
                        msg = "Can't find library file lib{}.dylib"
                        raise ValueError(msg.format(basename))
                assert fname.startswith('/')
                _flann_lib = os.path.abspath(fname)
                return _flann_lib
    elif sys.platform.startswith('linux'):
        out = subprocess.check_output(['ldd', so_name]).split('\n')
        for line in out:
            shortname, path = re.match(_ldd_re, line).groups()
            if 'libflann' in shortname and 'libflann_cpp' not in shortname:
                _flann_lib = os.path.abspath(path)
                return _flann_lib
    else:
        raise OSError("get_flann_lib doesn't know how to handle this OS")

def get_flann_include():
    '''
    Gets the location of the flann headers that users should link against.
    '''
    # TODO: do this a real way....
    import os
    return os.path.abspath(os.path.join(os.path.dirname(get_flann_lib()),
                                        '../include'))



import errno
from functools import partial
import os
import subprocess
import sys

try:
    import numpy
except ImportError as e:
    raise ImportError("cyflann requires numpy to be installed:\n{}".format(e))
    # Don't do this in the setup() requirements, because otherwise pip and
    # friends get too eager about updating numpy.

try:
    from setuptools import setup
    from setuptools.extension import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

################################################################################
# The following chunk of code for querying pkg-config is loosely based on code
# from the cffi package, which is licensed as follows:
#
#    The MIT License
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or
#    sell copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included
#    in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#    DEALINGS IN THE SOFTWARE.

def _ask_pkg_config(lib, option, result_prefix='', sysroot=False, default=None):
    pkg_config = os.environ.get('PKG_CONFIG', 'pkg-config')
    try:
        output = subprocess.check_output([pkg_config, option, lib])
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
    except subprocess.CalledProcessError:
        pass
    else:
        res = output.decode().strip().split()

        # '-I/usr/...' -> '/usr/...'
        for x in res:
            assert x.startswith(result_prefix)
        res = [x[len(result_prefix):] for x in res]

        if sysroot:
            sysroot = os.environ.get('PKG_CONFIG_SYSROOT_DIR', '')
        if sysroot:
            # old versions of pkg-config don't support this env var,
            # so here we emulate its effect if needed
            res = [path if path.startswith(sysroot) else sysroot + path
                   for path in res]

        return res
    return [] if default is None else default


def get_pkg_info(name):
    ask = partial(_ask_pkg_config, name)
    return {
        'libraries': ask('--libs-only-l', '-l', default=[name]),
        'include_dirs': ask('--cflags-only-I', '-I', sysroot=True),
        'library_dirs': ask('--libs-only-L', '-L', sysroot=True),
        'extra_compile_args': ask('--cflags-only-other'),
        'extra_link_args': ask('--libs-only-other'),
        'runtime_library_dirs': ask('--variable=libdir'),
    }

if os.environ.get('FLANN_DIR', False):
    pre = partial(os.path.join, os.environ['FLANN_DIR'])
    flann_info = {
        'libraries': ['flann', 'flann_cpp'],
        'include_dirs': [pre('include')],
        'library_dirs': [pre('lib')],
        'extra_compile_args': [],
        'extra_link_args': [],
        'runtime_library_dirs': [pre('lib')],
    }
else:
    flann_info = get_pkg_info('flann')

if sys.platform == 'darwin':
    # OSX rpath support is weird and stupid; just use an absolute path instead
    ld = flann_info['library_dirs']
    rld = flann_info['runtime_library_dirs']

    if len(ld) == 1 and rld == ld:
        del flann_info['runtime_library_dirs'], flann_info['library_dirs']
        flann_info['libraries'] = [
            os.path.join(ld[0], lib) for lib in flann_info['libraries']]
    elif not ld and not rld:
        pass
    else:
        import warnings
        msg = "Something unexpected is going on here. Good luck!"
        warnings.warn(msg)


################################################################################

try:
    from Cython.Build import cythonize
except ImportError:
    src_path = os.path.join(os.path.dirname(__file__), 'cyflann')
    if not os.path.exists(os.path.join(src_path, 'index.c')):
        msg = "index extension needs to be compiled but cython isn't available"
        raise ImportError(msg)
    ext_modules = [
        Extension("cyflann.index", ["cyflann/index.c"]),
    ]
else:
    ext_modules = cythonize("cyflann/*.pyx", "cyflann/*.pdx")

for ext in ext_modules:
    ext.__dict__.update(flann_info)


setup(
    name='cyflann',
    version='0.1.13',
    author='Dougal J. Sutherland',
    author_email='dougal@gmail.com',
    packages=['cyflann', 'cyflann.tests'],
    package_data={'cyflann': ['*.pyx', '*.pxd']},
    url='https://github.com/dougalsutherland/cyflann',
    description='A Cython-based interface to the FLANN nearest neighbors '
                'library.',
    long_description=open('README.rst').read(),
    license='BSD 3-clause',
    include_dirs=[numpy.get_include()],
    ext_modules=ext_modules,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Cython",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    zip_safe=False,  # not unsafe but no point, since it's just a c ext
)



from functools import partial
import numpy as np
import pyflann
import cyflann

def check_match(py_ids, py_dists, cy_ids, cy_dists):
    assert np.all(py_ids == cy_ids), \
           "{}/{} different".format(np.sum(py_ids != cy_ids), py_ids.size)
    assert np.allclose(py_dists, cy_dists, atol=1e-5, rtol=1e-4), \
           "max distance {}".format(np.abs(py_dists - cy_dists).max())

def test_normal():
    for dim in [1, 4, 10]:
        for k in [1, 2, 5]:
            data = np.random.normal(scale=100, size=(1000, dim))
            query = np.random.normal(scale=100, size=(100, dim))

            py = pyflann.FLANN(algorithm='kdtree_single')
            cy = cyflann.FLANNIndex(algorithm='kdtree_single')

            py.build_index(data)
            cy.build_index(data)

            py_ids, py_dists = py.nn_index(query, k)
            cy_ids, cy_dists = cy.nn_index(query, k)

            f = partial(check_match, py_ids, py_dists, cy_ids, cy_dists)
            f.description = 'normal vs pyflann - dim {} - k {}'.format(dim, k)
            yield f
